# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple test program for hid interface.

This simple test program lists all of the HID devices on the
local system.
"""
from __future__ import print_function

from pyu2f import hid


def main():
  devs = hid.Enumerate()
  for dev in devs:
    print(dev)


if __name__ == '__main__':
  main()
